"""Functions for the automated map generation for certain cases.
   ============================================================

   The functions are:

   o get_vector_field_mag_1
      : used for the mag 1 template (normal conducting magnet) :

   o get_vector_field_mag_2
      : used for the mag 2 template (superconducting magnet) :
   
   o get_vector_field_mag_3
      : used for the mag 1 template (normal conducting magnet) :

   o get_map_ncsc
      : used to simulate the cross talk between hadron absorber and
      superconducting magnet :
"""

import os
import numpy as np
import pandas as pd
import pyvista as pv
import time
import gmsh
import matplotlib.pyplot as plt
from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import cg
from scipy.sparse import diags
from scipy.interpolate import griddata 

import snoopy

def get_vector_field_mag_1(parameters, df_index=0, lc=0.2,
                  geo_th=1e-5, run_gmsh=False, plot_geo=False,
                  plot_result=False, result_directory='none', result_spec='',
                  eval_pos=np.zeros((0, 3))):
   '''Get the vector point cloud for the magnet 1 template.
    
   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param geo_th:
      A threshold for the identification of boundary surfaces.
      Adjust it if You generate very small features.

   :param run_gmsh:
      Set this flag to true if You like to run the gmsh gui after the mesh
      was generated.

   :param plot_geo:
      Set this flag to true if You like to generate a 3D plot of the geometry.

   :param plot_result:
      Set this flag to true if You like to generate a 3D plot of the result.

   :param result_directory:
      The result directory in case You like to store the solution somewhere.

   :param result_spec:
      A specifyer for the result files.

   :param eval_pos:
      Additional positions to be evaluated. Default empty.

   :return:
      The positions and field components in a 3D numpy grid.
   '''

   # launch time measurement
   t_s = time.time()

   # ====================================================
   # read the material data
   reluctance_iron = snoopy.Reluctance(parameters["material"][df_index])
   reluctance_air = snoopy.ConstantReluctance(1e7/4/np.pi)

   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 1 template")

   # we get the geometry parameters as variables for convinience
   X_mgap_1 = parameters["Xmgap1(m)"][df_index]
   X_mgap_2 = parameters["Xmgap2(m)"][df_index]

   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the iron domain
   vol_iron = snoopy.add_SHIP_iron_yoke(gmsh.model, X_mgap_1,
                                                    X_core_1,
                                                    X_void_1,
                                                    X_yoke_1,
                                                    X_mgap_2, 
                                                    X_core_2,
                                                    X_void_2,
                                                    X_yoke_2,
                                                    Y_core_1,
                                                    Y_void_1,
                                                    Y_yoke_1,
                                                    Y_core_2,
                                                    Y_void_2,
                                                    Y_yoke_2,
                                                    Z_len, 
                                                    Z_pos=Z_pos,
                                                    lc=lc,
                                                    lc_inner=0.2*lc,
                                                    yoke_type=1)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_iron)], [(3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_iron = fragments[0][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_iron], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   # we now need to collect all Dirichlet boundaries
   boundary_entities = gmsh.model.getEntities(2)

   # this list will store the boundary tags
   dirichlet_boundaries = []

   for i, be in boundary_entities:

      min_uv, max_uv = gmsh.model.getParametrizationBounds(2, be)
      u = 0.5*(max_uv[0] + min_uv[0])
      v = 0.5*(max_uv[1] + min_uv[1])

      coord = gmsh.model.getValue(2, be, [u, v])
      normal = gmsh.model.getNormal(be, [u, v])

      if (abs(coord[0] - 0.0) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[0] - lim_x) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[1] - lim_y) < geo_th and abs(abs(normal[1]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_min) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_max) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

   # add a physical group for this boundary condition
   gmsh.model.addPhysicalGroup(2, dirichlet_boundaries, 1, name = "Dirichlet Boundary")

   gmsh.model.occ.synchronize()

   if run_gmsh:
      gmsh.fltk.run()


   # ====================================================
   # Make the coil objects

   # this list stores the coil objects
   coil_list = []

   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   # these are the vertical positions
   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   if X_mgap_1 == 0.0 or X_mgap_2 == 0.0:

      # make only a single coil
      kp = np.array([[-X_core_2 - yoke_spacer - ins, Z_pos + Z_len             ],
                     [-X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                     [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                     [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [-X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [-X_core_1 - yoke_spacer - ins,   Z_pos                   ]])

      

      coil_list.append(snoopy.RacetrackCoil(kp, y, coil_radius, current/num_cond))

   else:

      # make two coils

      kp_1 = np.array([[ X_mgap_2 - yoke_spacer - ins, Z_pos + Z_len           ],
                     [ X_mgap_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                     [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                     [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [ X_mgap_1,                       Z_pos-yoke_spacer - ins ],
                     [ X_mgap_1 - yoke_spacer - ins,   Z_pos                   ]])
      
      kp_2 = kp_1.copy()
      kp_2[:, 0] *= -1.0
      
      coil_list.append(snoopy.RacetrackCoil(kp_1, y, coil_radius, current/num_cond))
      coil_list.append(snoopy.RacetrackCoil(kp_2, y, coil_radius, current/num_cond))

   if plot_geo:

      # ====================================================
      # Plot the magnet
      pl = pv.Plotter(shape=(1, 1), off_screen=False)

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
      for coil in coil_list:
         coil.plot_pv(pl)
      pl.show_grid()
      pl.add_axes()

      light_1 = pv.Light((1.5, 0.0, -1.2), (0, 0, 0), 'white')
      light_2 = pv.Light((-0.15, 0.2, 0.1), (0, 0, 0), 'white')


      pl.add_light(light_1)
      pl.add_light(light_2)

      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.show()
   

   # ====================================================
   # Solve the problem
   solver = snoopy.RedMVPSolver(gmsh.model, coil_list, 
                                [dom_iron], [dom_air],
                                [reluctance_iron],
                                quad_order=12, max_newton_iterations=150)
   
   x = solver.solve()


   # ====================================================
   # Get the point cloud
   points, B_i = solver.curl_curl_factory.compute_B(x, quad_order=field_density)

   B_coil = 0.0*B_i
   for coil in coil_list:
      B_coil += coil.compute_B(points)

   # stop time measurement
   t_e = time.time()

   print('elapsed time = {:.2f} sec'.format(t_e - t_s))

   if plot_result:
      # ====================================================
      # Plot the solution in 3D
      pl = pv.Plotter(shape=(1, 2))

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_i, title='B iron in T', mag=0.1, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()

      pl.subplot(0, 1)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_coil, title='B coil in T', mag=30.0, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()
      pl.show()

   # ====================================================
   # Store the solution
   if not result_directory == 'none':
      np.save(os.path.join(result_directory, 'x' + result_spec + '.npy'), x)
      
      out_params = parameters.iloc[df_index]
      out_params_df = out_params.to_frame().T
      out_params_df.to_csv(os.path.join(result_directory, 'parameters' + result_spec + '.csv'), index=False)


   ret_vals = [points, B_i + B_coil]

   if eval_pos.shape[0] > 0:
      
      # ====================================================
      # Evaluate at the given positions

      B_iron = snoopy.evaluate_curl_curl_solution(eval_pos, gmsh.model, [dom_iron, dom_air], x, solver.get_global_ids())

      B_coil = 0.0*B_iron
      for coil in coil_list:
         B_coil += coil.compute_B(eval_pos)
      
      B_tot = B_coil + B_iron 

      if not result_directory == 'none':
         output_filename = os.path.join(result_directory, 'B' + result_spec + '.csv')

      else:
         print('Warning, no result directory specified. I store the evaluations in snoopy directory.')
         output_filename = 'B' + result_spec + '.csv'

      out_df = pd.DataFrame(data=np.append(eval_pos, B_tot, axis=1),
                              columns=['x(m)', 'y(m)', 'z(m)', 'Bx(T)', 'By(T)', 'Bz(T)']).to_csv(output_filename, index=False)

   return ret_vals


def get_vector_field_mag_2(parameters, df_index=0, lc=0.4,
                  geo_th=1e-5, run_gmsh=False, plot_geo=False,
                  plot_result=False, result_directory='none', result_spec='',
                  eval_pos=np.zeros((0, 3))):
   '''Get the vector point cloud for the magnet 2 template.
    
   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param geo_th:
      A threshold for the identification of boundary surfaces.
      Adjust it if You generate very small features.

   :param run_gmsh:
      Set this flag to true if You like to run the gmsh gui after the mesh
      was generated.

   :param plot_geo:
      Set this flag to true if You like to generate a 3D plot of the geometry.

   :param plot_result:
      Set this flag to true if You like to generate a 3D plot of the result.

   :param result_directory:
      The result directory in case You like to store the solution somewhere.

   :param result_spec:
      A specifyer for the result files.

   :param eval_pos:
      Additional positions to be evaluated. Default empty.

   :return:
      The positions and field components in a 3D numpy grid.
   '''

   # launch time measurement
   t_s = time.time()

   # ====================================================
   # read the material data
   reluctance_iron = snoopy.Reluctance(parameters["material"][df_index])
   reluctance_air = snoopy.ConstantReluctance(1e7/4/np.pi)

   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 2 template")

   # we get the geometry parameters as variables for convinience
   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the core domain
   vol_core = snoopy.add_SHIP_iron_core(gmsh.model, X_core_1,
                                                    X_core_2,
                                                    Y_core_1,
                                                    Y_core_2,
                                                    Z_len,
                                                    Z_pos=Z_pos,
                                                    lc=0.3*lc)

   # the yoke domain
   vol_yoke = snoopy.add_SHIP_iron_yoke(gmsh.model, 0.0, X_core_1,
                                                         X_void_1,
                                                         X_yoke_1,
                                                         0.0, 
                                                         X_core_2,
                                                         X_void_2,
                                                         X_yoke_2,
                                                         Y_core_1,
                                                         Y_void_1,
                                                         Y_yoke_1,
                                                         Y_core_2,
                                                         Y_void_2,
                                                         Y_yoke_2,
                                                         Z_len, 
                                                         Z_pos=Z_pos,
                                                         lc=lc,
                                                         lc_inner=0.3*lc,
                                                         yoke_type=2)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_core)], [(3, vol_yoke), (3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_core = fragments[0][1]
   dom_yoke = fragments[2][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_core, dom_yoke], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   # we now need to collect all Dirichlet boundaries
   boundary_entities = gmsh.model.getEntities(2)

   # this list will store the boundary tags
   dirichlet_boundaries = []

   for i, be in boundary_entities:

      min_uv, max_uv = gmsh.model.getParametrizationBounds(2, be)
      u = 0.5*(max_uv[0] + min_uv[0])
      v = 0.5*(max_uv[1] + min_uv[1])

      coord = gmsh.model.getValue(2, be, [u, v])
      normal = gmsh.model.getNormal(be, [u, v])

      if (abs(coord[0] - 0.0) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[0] - lim_x) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[1] - lim_y) < geo_th and abs(abs(normal[1]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_min) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_max) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

   # add a physical group for this boundary condition
   gmsh.model.addPhysicalGroup(2, dirichlet_boundaries, 1, name = "Dirichlet Boundary")

   gmsh.model.occ.synchronize()

   if run_gmsh:
      gmsh.fltk.run()


   # ====================================================
   # Make a coil object
   kp = np.array([[-X_core_2 - yoke_spacer - ins, Z_pos + Z_len             ],
                  [-X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                  [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                  [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                  [-X_core_1,                       Z_pos-yoke_spacer - ins ],
                  [-X_core_1 - yoke_spacer - ins,   Z_pos                   ]])
   
   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   coil = snoopy.RacetrackCoil(kp, y, coil_radius, current/num_cond)

   if plot_geo:

      # ====================================================
      # Plot the magnet
      pl = pv.Plotter(shape=(1, 1), off_screen=False)

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_core, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_yoke, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
      coil.plot_pv(pl)
      pl.show_grid()
      pl.add_axes()

      light_1 = pv.Light((1.5, 0.0, -1.2), (0, 0, 0), 'white')
      light_2 = pv.Light((-0.15, 0.2, 0.1), (0, 0, 0), 'white')


      pl.add_light(light_1)
      pl.add_light(light_2)

      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.show()
   
   # ====================================================
   # Solve the problem
   solver = snoopy.RedMVPSolver(gmsh.model, [coil], 
                                [dom_core, dom_yoke], [dom_air],
                                [reluctance_iron, reluctance_iron],
                                quad_order=10, max_newton_iterations=25)
   
   x = solver.solve()

   # ====================================================
   # Get the point cloud
   points, B_i = solver.curl_curl_factory.compute_B(x, quad_order=field_density)

   B_coil = coil.compute_B(points)

   # stop time measurement
   t_e = time.time()

   print('elapsed time = {:.2f} sec'.format(t_e - t_s))

   if plot_result:
      # ====================================================
      # Plot the solution in 3D
      pl = pv.Plotter(shape=(1, 2))

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_core, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_yoke, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_i, title='B iron in T', mag=0.05, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()

      pl.subplot(0, 1)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_core, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_yoke, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_coil, title='B coil in T', mag=0.05, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()
      pl.show()

   # ====================================================
   # Store the solution
   if not result_directory == 'none':
      np.save(os.path.join(result_directory, 'x' + result_spec + '.npy'), x)
      
      out_params = parameters.iloc[df_index]
      out_params_df = out_params.to_frame().T
      out_params_df.to_csv(os.path.join(result_directory, 'parameters' + result_spec + '.csv'), index=False)


   ret_vals = [points, B_i + B_coil]

   if eval_pos.shape[0] > 0:
      
      # ====================================================
      # Evaluate at the given positions

      B_iron = snoopy.evaluate_curl_curl_solution(eval_pos, gmsh.model, [dom_core, dom_yoke, dom_air], x, solver.get_global_ids())
      B_coil = coil.compute_B(eval_pos)
      
      B_tot = B_coil + B_iron 

      if not result_directory == 'none':
         output_filename = os.path.join(result_directory, 'B' + result_spec + '.csv')

      else:
         print('Warning, no result directory specified. I store the evaluations in snoopy directory.')
         output_filename = 'B' + result_spec + '.csv'

      out_df = pd.DataFrame(data=np.append(eval_pos, B_tot, axis=1),
                              columns=['x(m)', 'y(m)', 'z(m)', 'Bx(T)', 'By(T)', 'Bz(T)']).to_csv(output_filename, index=False)

   return ret_vals


def get_vector_field_mag_3(parameters, df_index=0, lc=0.4,
                  geo_th=1e-5, run_gmsh=False, plot_geo=False,
                  plot_result=False, result_directory='none', result_spec='',
                  eval_pos=np.zeros((0, 3))):
   '''Get the vector point cloud for the magnet 3 template.
    
   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param geo_th:
      A threshold for the identification of boundary surfaces.
      Adjust it if You generate very small features.

   :param run_gmsh:
      Set this flag to true if You like to run the gmsh gui after the mesh
      was generated.

   :param plot_geo:
      Set this flag to true if You like to generate a 3D plot of the geometry.

   :param plot_result:
      Set this flag to true if You like to generate a 3D plot of the result.

   :param result_directory:
      The result directory in case You like to store the solution somewhere.

   :param result_spec:
      A specifyer for the result files.

   :param eval_pos:
      Additional positions to be evaluated. Default empty.

   :return:
      The positions and field components in a 3D numpy grid.
   '''


   # ====================================================
   # read the material data
   reluctance_iron = snoopy.Reluctance(parameters["material"][df_index])
   reluctance_air = snoopy.ConstantReluctance(1e7/4/np.pi)

   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 3 template")

   # we get the geometry parameters as variables for convinience
   X_mgap_1 = parameters["Xmgap1(m)"][df_index]
   X_mgap_2 = parameters["Xmgap2(m)"][df_index]

   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the iron domain
   vol_iron = snoopy.add_SHIP_iron_yoke(gmsh.model, X_mgap_1,
                                                    X_core_1,
                                                    X_void_1,
                                                    X_yoke_1,
                                                    X_mgap_2, 
                                                    X_core_2,
                                                    X_void_2,
                                                    X_yoke_2,
                                                    Y_core_1,
                                                    Y_void_1,
                                                    Y_yoke_1,
                                                    Y_core_2,
                                                    Y_void_2,
                                                    Y_yoke_2,
                                                    Z_len, 
                                                    Z_pos=Z_pos,
                                                    lc=lc,
                                                    lc_inner=0.3*lc,
                                                    yoke_type=3)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_iron)], [(3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_iron = fragments[0][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_iron], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   # we now need to collect all Dirichlet boundaries
   boundary_entities = gmsh.model.getEntities(2)

   # this list will store the boundary tags
   dirichlet_boundaries = []

   for i, be in boundary_entities:

      min_uv, max_uv = gmsh.model.getParametrizationBounds(2, be)
      u = 0.5*(max_uv[0] + min_uv[0])
      v = 0.5*(max_uv[1] + min_uv[1])

      coord = gmsh.model.getValue(2, be, [u, v])
      normal = gmsh.model.getNormal(be, [u, v])

      if (abs(coord[0] - 0.0) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[0] - lim_x) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[1] - lim_y) < geo_th and abs(abs(normal[1]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_min) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_max) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

   # add a physical group for this boundary condition
   gmsh.model.addPhysicalGroup(2, dirichlet_boundaries, 1, name = "Dirichlet Boundary")

   gmsh.model.occ.synchronize()

   if run_gmsh:
      gmsh.fltk.run()


   # ====================================================
   # Make the coil objects

   # this list stores the coil objects
   coil_list = []

   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   # these are the vertical positions
   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   # make two coils

   kp_1 = np.array([[ X_void_2 - yoke_spacer - ins, Z_pos + Z_len           ],
                  [ X_void_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_yoke_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_yoke_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                  [ X_yoke_1 + yoke_spacer + ins,   Z_pos                   ],
                  [ X_yoke_1,                       Z_pos-yoke_spacer - ins ],
                  [ X_void_1,                       Z_pos-yoke_spacer - ins ],
                  [ X_void_1 - yoke_spacer - ins,   Z_pos                   ]])
      
   kp_2 = kp_1.copy()
   kp_2[:, 0] *= -1.0
      
   coil_list.append(snoopy.RacetrackCoil(kp_1, y, coil_radius, current/num_cond))
   coil_list.append(snoopy.RacetrackCoil(kp_2, y, coil_radius, current/num_cond))

   if plot_geo:

      # ====================================================
      # Plot the magnet
      pl = pv.Plotter(shape=(1, 1), off_screen=False)

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
      for coil in coil_list:
         coil.plot_pv(pl)
      pl.show_grid()
      pl.add_axes()

      light_1 = pv.Light((1.5, 0.0, -1.2), (0, 0, 0), 'white')
      light_2 = pv.Light((-0.15, 0.2, 0.1), (0, 0, 0), 'white')


      pl.add_light(light_1)
      pl.add_light(light_2)

      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.show()
   

   # ====================================================
   # Solve the problem
   solver = snoopy.RedMVPSolver(gmsh.model, coil_list, 
                                [dom_iron], [dom_air],
                                [reluctance_iron],
                                quad_order=12, max_newton_iterations=50)
   
   x = solver.solve()


   # ====================================================
   # Get the point cloud
   points, B_i = solver.curl_curl_factory.compute_B(x, quad_order=field_density)

   B_coil = 0.0*B_i
   for coil in coil_list:
      B_coil += coil.compute_B(points)


   if plot_result:
      # ====================================================
      # Plot the solution in 3D
      pl = pv.Plotter(shape=(1, 2))

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_i, title='B iron in T', mag=0.1, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()

      pl.subplot(0, 1)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_coil, title='B coil in T', mag=30.0, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()
      pl.show()

   # ====================================================
   # Store the solution
   if not result_directory == 'none':
      np.save(os.path.join(result_directory, 'x' + result_spec + '.npy'), x)
      
      out_params = parameters.iloc[df_index]
      out_params_df = out_params.to_frame().T
      out_params_df.to_csv(os.path.join(result_directory, 'parameters' + result_spec + '.csv'), index=False)


   ret_vals = [points, B_i + B_coil]

   if eval_pos.shape[0] > 0:
      
      # ====================================================
      # Evaluate at the given positions

      B_iron = snoopy.evaluate_curl_curl_solution(eval_pos, gmsh.model, [dom_iron, dom_air], x, solver.get_global_ids())

      B_coil = 0.0*B_iron
      for coil in coil_list:
         B_coil += coil.compute_B(eval_pos)
      
      B_tot = B_coil + B_iron 

      if not result_directory == 'none':
         output_filename = os.path.join(result_directory, 'B' + result_spec + '.csv')

      else:
         print('Warning, no result directory specified. I store the evaluations in snoopy directory.')
         output_filename = 'B' + result_spec + '.csv'

      out_df = pd.DataFrame(data=np.append(eval_pos, B_tot, axis=1),
                              columns=['x(m)', 'y(m)', 'z(m)', 'Bx(T)', 'By(T)', 'Bz(T)']).to_csv(output_filename, index=False)

   return ret_vals



def get_vector_field_ncsc(parameters, df_index=0, lc=0.4,
                  geo_th=1e-5, run_gmsh=False, plot_geo=False,
                  plot_result=False, result_directory='none', result_spec='',
                  eval_pos=np.zeros((0, 3))):
   '''Get the vector point cloud for the HASC template, i.e. the coupling
   of hardon absorber and superconducting magnet.
    
   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param geo_th:
      A threshold for the identification of boundary surfaces.
      Adjust it if You generate very small features.

   :param run_gmsh:
      Set this flag to true if You like to run the gmsh gui after the mesh
      was generated.

   :param plot_geo:
      Set this flag to true if You like to generate a 3D plot of the geometry.

   :param plot_result:
      Set this flag to true if You like to generate a 3D plot of the result.

   :param result_directory:
      The result directory in case You like to store the solution somewhere.

   :param result_spec:
      A specifyer for the result files.

   :param eval_pos:
      Additional positions to be evaluated. Default empty.

   :return:
      The positions and field components in a 3D numpy grid.
   '''

   # launch time measurement
   t_s = time.time()

   # ====================================================
   # read the material data
   reluctance_iron = snoopy.Reluctance(parameters["material"][df_index])
   reluctance_air = snoopy.ConstantReluctance(1e7/4/np.pi)

   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 1 template")

   # Read the geometry parameters for the normal conducting magnet
   X_mgap_1_1 = parameters["Xmgap1(m)"][df_index]
   X_mgap_2_1 = parameters["Xmgap2(m)"][df_index]

   X_core_1_1 = parameters["Xcore1(m)"][df_index]
   X_core_2_1 = parameters["Xcore2(m)"][df_index]

   X_void_1_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2_1 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2_1 = parameters["Xyoke2(m)"][df_index]

   Y_core_1_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2_1 = parameters["Ycore2(m)"][df_index]

   Y_void_1_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2_1 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2_1 = parameters["Yyoke2(m)"][df_index]

   Z_len_1 = parameters["Z_len(m)"][df_index]
   Z_pos_1 = parameters["Z_pos(m)"][df_index]
   
   yoke_spacer_1 = parameters["yoke_spacer(m)"][df_index]
   ins_1 = parameters["insulation(m)"][df_index]
   current_1 = parameters["NI(A)"][df_index]
   coil_radius_1 = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   delta_x_1 = parameters["delta_x(m)"][df_index]
   delta_y_1 = parameters["delta_y(m)"][df_index]
   delta_z_1 = parameters["delta_z(m)"][df_index]

   field_density_1 = 0.5*parameters["field_density"][df_index]

   # Read the geometry parameters for the superconducting magnet
   X_mgap_1_2 = parameters["Xmgap1(m)"][df_index + 1]
   X_mgap_2_2 = parameters["Xmgap2(m)"][df_index + 1]

   X_core_1_2 = parameters["Xcore1(m)"][df_index + 1]
   X_core_2_2 = parameters["Xcore2(m)"][df_index + 1]

   X_void_1_2 = parameters["Xvoid1(m)"][df_index + 1]
   X_void_2_2 = parameters["Xvoid2(m)"][df_index + 1]

   X_yoke_1_2 = parameters["Xyoke1(m)"][df_index + 1]
   X_yoke_2_2 = parameters["Xyoke2(m)"][df_index + 1]

   Y_core_1_2 = parameters["Ycore1(m)"][df_index + 1]
   Y_core_2_2 = parameters["Ycore2(m)"][df_index + 1]

   Y_void_1_2 = parameters["Yvoid1(m)"][df_index + 1]
   Y_void_2_2 = parameters["Yvoid2(m)"][df_index + 1]

   Y_yoke_1_2 = parameters["Yyoke1(m)"][df_index + 1]
   Y_yoke_2_2 = parameters["Yyoke2(m)"][df_index + 1]

   Z_len_2 = parameters["Z_len(m)"][df_index + 1]
   Z_pos_2 = parameters["Z_pos(m)"][df_index + 1]
   
   yoke_spacer_2 = parameters["yoke_spacer(m)"][df_index + 1]
   ins_2 = parameters["insulation(m)"][df_index + 1]
   current_2 = parameters["NI(A)"][df_index + 1]
   coil_radius_2 = 0.5*parameters["coil_diam(mm)"][df_index + 1]*1e-3

   delta_x_2 = parameters["delta_x(m)"][df_index + 1]
   delta_y_2 = parameters["delta_y(m)"][df_index + 1]
   delta_z_2 = parameters["delta_z(m)"][df_index + 1]

   field_density_2 = 0.5*parameters["field_density"][df_index + 1]

   # the limits in x, y, and z
   lim_x = max([max([X_yoke_1_1, X_yoke_2_1]) + delta_x_1, max([X_yoke_1_2, X_yoke_2_2]) + delta_x_2])
   lim_y = max([max([Y_yoke_1_1, Y_yoke_2_1]) + delta_y_1, max([Y_yoke_1_2, Y_yoke_2_2]) + delta_y_2])
   z_min = Z_pos_1 - delta_z_1
   z_max = Z_pos_2 + Z_len_2 + delta_z_2
   
   # the nc magnet iron domain
   vol_nc = snoopy.add_SHIP_iron_yoke(gmsh.model, X_mgap_1_1,
                                                    X_core_1_1,
                                                    X_void_1_1,
                                                    X_yoke_1_1,
                                                    X_mgap_2_1, 
                                                    X_core_2_1,
                                                    X_void_2_1,
                                                    X_yoke_2_1,
                                                    Y_core_1_1,
                                                    Y_void_1_1,
                                                    Y_yoke_1_1,
                                                    Y_core_2_1,
                                                    Y_void_2_1,
                                                    Y_yoke_2_1,
                                                    Z_len_1, 
                                                    Z_pos=Z_pos_1,
                                                    lc=lc,
                                                    lc_inner=lc,
                                                    yoke_type=1)

   # the sc magnet core domain
   vol_sc_core = snoopy.add_SHIP_iron_core(gmsh.model, X_core_1_2,
                                                    X_core_2_2,
                                                    Y_core_1_2,
                                                    Y_core_2_2,
                                                    Z_len_2,
                                                    Z_pos=Z_pos_2,
                                                    lc=lc)

   # the sc magnet yoke domain
   vol_sc_yoke = snoopy.add_SHIP_iron_yoke(gmsh.model, 0.0, X_core_1_2,
                                                         X_void_1_2,
                                                         X_yoke_1_2,
                                                         0.0, 
                                                         X_core_2_2,
                                                         X_void_2_2,
                                                         X_yoke_2_2,
                                                         Y_core_1_2,
                                                         Y_void_1_2,
                                                         Y_yoke_1_2,
                                                         Y_core_2_2,
                                                         Y_void_2_2,
                                                         Y_yoke_2_2,
                                                         Z_len_2, 
                                                         Z_pos=Z_pos_2,
                                                         lc=lc,
                                                         lc_inner=lc,
                                                         yoke_type=2)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_air)], [(3, vol_nc), (3, vol_sc_core), (3, vol_sc_yoke)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_nc = fragments[0][1]
   dom_sc_core = fragments[1][1]
   dom_air = fragments[2][1]
   dom_sc_yoke = fragments[3][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_nc, dom_sc_core, dom_sc_yoke], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   gmsh.option.setNumber("Mesh.MeshSizeFactor", lc)

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   # we now need to collect all Dirichlet boundaries
   boundary_entities = gmsh.model.getEntities(2)

   # this list will store the boundary tags
   dirichlet_boundaries = []

   for i, be in boundary_entities:

      min_uv, max_uv = gmsh.model.getParametrizationBounds(2, be)
      u = 0.5*(max_uv[0] + min_uv[0])
      v = 0.5*(max_uv[1] + min_uv[1])

      coord = gmsh.model.getValue(2, be, [u, v])
      normal = gmsh.model.getNormal(be, [u, v])

      if (abs(coord[0] - 0.0) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[0] - lim_x) < geo_th and abs(abs(normal[0]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[1] - lim_y) < geo_th and abs(abs(normal[1]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_min) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

      elif (abs(coord[2] - z_max) < geo_th and abs(abs(normal[2]) - 1.0) < geo_th):
         dirichlet_boundaries.append(be)

   # add a physical group for this boundary condition
   gmsh.model.addPhysicalGroup(2, dirichlet_boundaries, 1, name = "Dirichlet Boundary")

   gmsh.model.occ.synchronize()

   if run_gmsh:
      gmsh.fltk.run()


   # ====================================================
   # Make the coil objects

   # this list stores the coil objects
   coil_list = []

   # determine the slot sizes
   slot_size_1 = 2*min(Y_core_1_1, Y_core_2_1)
   slot_size_2 = 2*min(Y_core_1_2, Y_core_2_2)

   # determine the number of conductors
   num_cond_1 = np.int32(slot_size_1/2/(coil_radius_1 + ins_1))
   num_cond_2 = np.int32(slot_size_2/2/(coil_radius_2 + ins_2))

   # these are the vertical positions
   y_1 = np.linspace(-0.5*slot_size_1 + coil_radius_1 + ins_1,
                    0.5*slot_size_1 - coil_radius_1 - ins_1, num_cond_1)
   y_2 = np.linspace(-0.5*slot_size_2 + coil_radius_2 + ins_2,
                    0.5*slot_size_2 - coil_radius_2 - ins_2, num_cond_2)

   # make only a single coil for the first nc magnet
   kp_1 = np.array([[-X_core_2_1 - yoke_spacer_1 - ins_1, Z_pos_1 + Z_len_1             ],
                     [-X_core_2_1,          Z_pos_1 + Z_len_1 + yoke_spacer_1 + ins_1    ],
                     [ X_core_2_1,          Z_pos_1 + Z_len_1 + yoke_spacer_1 + ins_1    ],
                     [ X_core_2_1 + yoke_spacer_1 + ins_1,   Z_pos_1 + Z_len_1           ],
                     [ X_core_1_1 + yoke_spacer_1 + ins_1,   Z_pos_1                   ],
                     [ X_core_1_1,                       Z_pos_1-yoke_spacer_1 - ins_1 ],
                     [-X_core_1_1,                       Z_pos_1-yoke_spacer_1 - ins_1 ],
                     [-X_core_1_1 - yoke_spacer_1 - ins_1,   Z_pos_1                   ]])

   coil_list.append(snoopy.RacetrackCoil(kp_1, y_1, coil_radius_1, current_1/num_cond_1))

   # make only a single coil for the next sc magnet
   kp_2 = np.array([[-X_core_2_2 - yoke_spacer_2 - ins_2, Z_pos_2 + Z_len_2             ],
                     [-X_core_2_2,          Z_pos_2 + Z_len_2 + yoke_spacer_2 + ins_2    ],
                     [ X_core_2_2,          Z_pos_2 + Z_len_2 + yoke_spacer_2 + ins_2    ],
                     [ X_core_2_2 + yoke_spacer_2 + ins_2,   Z_pos_2 + Z_len_2           ],
                     [ X_core_1_2 + yoke_spacer_2 + ins_2,   Z_pos_2                   ],
                     [ X_core_1_2,                       Z_pos_2-yoke_spacer_2 - ins_2 ],
                     [-X_core_1_2,                       Z_pos_2-yoke_spacer_2 - ins_2 ],
                     [-X_core_1_2 - yoke_spacer_2 - ins_2,   Z_pos_2                   ]]) 

   coil_list.append(snoopy.RacetrackCoil(kp_2, y_2, coil_radius_2, current_2/num_cond_2))


   if plot_geo:

      # ====================================================
      # Plot the magnet
      pl = pv.Plotter(shape=(1, 1), off_screen=False)

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_nc, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_core, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_yoke, reflect_xz=True, reflect_yz=True, show_edges=True, opacity=1.0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
      for coil in coil_list:
         coil.plot_pv(pl)
      pl.show_grid()
      pl.add_axes()

      light_1 = pv.Light((1.5, 0.0, -1.2), (0, 0, 0), 'white')
      light_2 = pv.Light((-0.15, 0.2, 0.1), (0, 0, 0), 'white')


      pl.add_light(light_1)
      pl.add_light(light_2)

      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.show()
   

   # ====================================================
   # Solve the problem
   solver = snoopy.RedMVPSolver(gmsh.model, coil_list, 
                                [dom_nc, dom_sc_core, dom_sc_yoke], [dom_air],
                                [reluctance_iron, reluctance_iron, reluctance_iron],
                                quad_order=12, max_newton_iterations=50)
   
   x = solver.solve()


   # ====================================================
   # Get the point cloud
   
   field_density = max([field_density_1, field_density_2])

   points, B_i = solver.curl_curl_factory.compute_B(x, quad_order=field_density)

   B_coil = 0.0*B_i
   for coil in coil_list:
      B_coil += coil.compute_B(points)

   # stop time measurement
   t_e = time.time()

   print('elapsed time = {:.2f} sec'.format(t_e - t_s))

   if plot_result:
      # ====================================================
      # Plot the solution in 3D
      pl = pv.Plotter(shape=(1, 2))

      pl.subplot(0, 0)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_nc, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_core, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_yoke, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_i, title='B iron in T', mag=0.1, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()

      pl.subplot(0, 1)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_nc, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_core, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_sc_yoke, plot_volume=False)
      snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, plot_volume=False, opacity=0.0)
      snoopy.plot_vector_field(pl, points, B_coil, title='B coil in T', mag=0.1, sym_yz=1, sym_xz=2, opacity='linear')
      pl.show_grid()
      pl.camera_position = 'zy'
      pl.camera.elevation = 35
      pl.camera.azimuth = 45
      pl.add_axes()
      pl.show()

   # ====================================================
   # Store the solution
   if not result_directory == 'none':
      np.save(os.path.join(result_directory, 'x' + result_spec + '.npy'), x)
      
      out_params = parameters.iloc[df_index]
      out_params_df = out_params.to_frame().T
      out_params_df.to_csv(os.path.join(result_directory, 'parameters' + result_spec + '.csv'), index=False)


   ret_vals = [points, B_i + B_coil]

   if eval_pos.shape[0] > 0:
      
      # ====================================================
      # Evaluate at the given positions

      B_iron = snoopy.evaluate_curl_curl_solution(eval_pos, gmsh.model, [dom_nc, dom_sc_core, dom_sc_yoke, dom_air], x, solver.get_global_ids())

      B_coil = 0.0*B_iron
      for coil in coil_list:
         B_coil += coil.compute_B(eval_pos)
      
      B_tot = B_coil + B_iron 

      if not result_directory == 'none':
         output_filename = os.path.join(result_directory, 'B' + result_spec + '.csv')

      else:
         print('Warning, no result directory specified. I store the evaluations in snoopy directory.')
         output_filename = 'B' + result_spec + '.csv'

      out_df = pd.DataFrame(data=np.append(eval_pos, B_tot, axis=1),
                              columns=['x(m)', 'y(m)', 'z(m)', 'Bx(T)', 'By(T)', 'Bz(T)']).to_csv(output_filename, index=False)

   return ret_vals

def plot_geometry_mag_1(pl, parameters, df_index=0, lc=1.0, opacity=0.0, show_edges=False, plot_feature_edges=True):
   '''Plot the geometry for the magnet 1 template.
    
   :params pl:
      The plotter object.

   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param opacity:
      The opacity of the iron domain. Default=False.

   :param show_edges:
      Set this flag to enable plotting the edges. Default=True.

   :return:
      None
   '''

   # ====================================================
   # read the material data
   reluctance_iron = snoopy.Reluctance(parameters["material"][df_index])
   reluctance_air = snoopy.ConstantReluctance(1e7/4/np.pi)

   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 1 template")

   # we get the geometry parameters as variables for convinience
   X_mgap_1 = parameters["Xmgap1(m)"][df_index]
   X_mgap_2 = parameters["Xmgap2(m)"][df_index]

   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the iron domain
   vol_iron = snoopy.add_SHIP_iron_yoke(gmsh.model, X_mgap_1,
                                                    X_core_1,
                                                    X_void_1,
                                                    X_yoke_1,
                                                    X_mgap_2, 
                                                    X_core_2,
                                                    X_void_2,
                                                    X_yoke_2,
                                                    Y_core_1,
                                                    Y_void_1,
                                                    Y_yoke_1,
                                                    Y_core_2,
                                                    Y_void_2,
                                                    Y_yoke_2,
                                                    Z_len, 
                                                    Z_pos=Z_pos,
                                                    lc=lc,
                                                    lc_inner=0.2*lc,
                                                    yoke_type=1)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_iron)], [(3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_iron = fragments[0][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_iron], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   gmsh.model.occ.synchronize()

   # ====================================================
   # Make the coil objects

   # this list stores the coil objects
   coil_list = []

   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   # these are the vertical positions
   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   if X_mgap_1 == 0.0 or X_mgap_2 == 0.0:

      # make only a single coil
      kp = np.array([[-X_core_2 - yoke_spacer - ins, Z_pos + Z_len             ],
                     [-X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                     [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                     [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [-X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [-X_core_1 - yoke_spacer - ins,   Z_pos                   ]])

      

      coil_list.append(snoopy.RacetrackCoil(kp, y, coil_radius, current/num_cond))

   else:

      # make two coils

      kp_1 = np.array([[ X_mgap_2 - yoke_spacer - ins, Z_pos + Z_len           ],
                     [ X_mgap_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                     [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                     [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                     [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                     [ X_mgap_1,                       Z_pos-yoke_spacer - ins ],
                     [ X_mgap_1 - yoke_spacer - ins,   Z_pos                   ]])
      
      kp_2 = kp_1.copy()
      kp_2[:, 0] *= -1.0
      
      coil_list.append(snoopy.RacetrackCoil(kp_1, y, coil_radius, current/num_cond))
      coil_list.append(snoopy.RacetrackCoil(kp_2, y, coil_radius, current/num_cond))

   snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, reflect_xz=True, reflect_yz=True, show_edges=show_edges, opacity=opacity, plot_feature_edges=plot_feature_edges)
   # snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
   for coil in coil_list:
      coil.plot_pv(pl)

   return

def plot_geometry_mag_2(pl, parameters, df_index=0, lc=1.0, opacity=0.0, show_edges=False, plot_feature_edges=True):
   '''Plot the geometry for the magnet 2 template.
    
   :params pl:
      The plotter object.

   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param opacity:
      The opacity of the iron domain. Default=False.

   :param show_edges:
      Set this flag to enable plotting the edges. Default=True.

   :return:
      None
   '''


   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 2 template")

   # we get the geometry parameters as variables for convinience
   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the core domain
   vol_core = snoopy.add_SHIP_iron_core(gmsh.model, X_core_1,
                                                    X_core_2,
                                                    Y_core_1,
                                                    Y_core_2,
                                                    Z_len,
                                                    Z_pos=Z_pos,
                                                    lc=0.3*lc)

   # the yoke domain
   vol_yoke = snoopy.add_SHIP_iron_yoke(gmsh.model, 0.0, X_core_1,
                                                         X_void_1,
                                                         X_yoke_1,
                                                         0.0, 
                                                         X_core_2,
                                                         X_void_2,
                                                         X_yoke_2,
                                                         Y_core_1,
                                                         Y_void_1,
                                                         Y_yoke_1,
                                                         Y_core_2,
                                                         Y_void_2,
                                                         Y_yoke_2,
                                                         Z_len, 
                                                         Z_pos=Z_pos,
                                                         lc=lc,
                                                         lc_inner=0.3*lc,
                                                         yoke_type=2)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_core)], [(3, vol_yoke), (3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_core = fragments[0][1]
   dom_yoke = fragments[2][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_core, dom_yoke], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   gmsh.model.occ.synchronize()


   # ====================================================
   # Make a coil object
   kp = np.array([[-X_core_2 - yoke_spacer - ins, Z_pos + Z_len             ],
                  [-X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_core_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_core_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                  [ X_core_1 + yoke_spacer + ins,   Z_pos                   ],
                  [ X_core_1,                       Z_pos-yoke_spacer - ins ],
                  [-X_core_1,                       Z_pos-yoke_spacer - ins ],
                  [-X_core_1 - yoke_spacer - ins,   Z_pos                   ]])
   
   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   coil = snoopy.RacetrackCoil(kp, y, coil_radius, current/num_cond)

   snoopy.plot_domain(pl, gmsh.model.mesh, dom_core, reflect_xz=True, reflect_yz=True, show_edges=show_edges, opacity=opacity, plot_feature_edges=plot_feature_edges)
   snoopy.plot_domain(pl, gmsh.model.mesh, dom_yoke, reflect_xz=True, reflect_yz=True, show_edges=show_edges, opacity=opacity, plot_feature_edges=plot_feature_edges)
   # snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
   coil.plot_pv(pl)

   return None

def plot_geometry_mag_3(pl, parameters, df_index=0, lc=1.0, opacity=0.0, show_edges=False, plot_feature_edges=True):
   '''Plot the geometry for the magnet 2 template.
    
   :params pl:
      The plotter object.

   :params parameters:
      The magnet parameters as pandas dataframe. See the
      SHiP documentation for details.

   :param df_index:
      The row in the pandas parameter dataframe.
   
   :param lc:
      The mesh size parameter. Default = 0.5.

   :param opacity:
      The opacity of the iron domain. Default=False.

   :param show_edges:
      Set this flag to enable plotting the edges. Default=True.

   :return:
      None
   '''
   # ====================================================
   # mesh generation
   gmsh.initialize()
   gmsh.model.add("make mesh mag 3 template")

   # we get the geometry parameters as variables for convinience
   X_mgap_1 = parameters["Xmgap1(m)"][df_index]
   X_mgap_2 = parameters["Xmgap2(m)"][df_index]

   X_core_1 = parameters["Xcore1(m)"][df_index]
   X_core_2 = parameters["Xcore2(m)"][df_index]

   X_void_1 = parameters["Xvoid1(m)"][df_index]
   X_void_2 = parameters["Xvoid2(m)"][df_index]

   X_yoke_1 = parameters["Xyoke1(m)"][df_index]
   X_yoke_2 = parameters["Xyoke2(m)"][df_index]

   Y_core_1 = parameters["Ycore1(m)"][df_index]
   Y_core_2 = parameters["Ycore2(m)"][df_index]

   Y_void_1 = parameters["Yvoid1(m)"][df_index]
   Y_void_2 = parameters["Yvoid2(m)"][df_index]

   Y_yoke_1 = parameters["Yyoke1(m)"][df_index]
   Y_yoke_2 = parameters["Yyoke2(m)"][df_index]

   Z_len = parameters["Z_len(m)"][df_index]
   Z_pos = parameters["Z_pos(m)"][df_index]

   delta_x = parameters["delta_x(m)"][df_index]
   delta_y = parameters["delta_y(m)"][df_index]
   delta_z = parameters["delta_z(m)"][df_index]
   
   yoke_spacer = parameters["yoke_spacer(m)"][df_index]

   ins = parameters["insulation(m)"][df_index]

   current = parameters["NI(A)"][df_index]

   coil_radius = 0.5*parameters["coil_diam(mm)"][df_index]*1e-3

   field_density = 0.5*parameters["field_density"][df_index]

   # the limits in x, y, and z
   lim_x = max([X_yoke_1, X_yoke_2]) + delta_x
   lim_y = max([Y_yoke_1, Y_yoke_2]) + delta_y
   z_min = Z_pos - delta_z
   z_max = Z_pos + Z_len + delta_z
   
   # the iron domain
   vol_iron = snoopy.add_SHIP_iron_yoke(gmsh.model, X_mgap_1,
                                                    X_core_1,
                                                    X_void_1,
                                                    X_yoke_1,
                                                    X_mgap_2, 
                                                    X_core_2,
                                                    X_void_2,
                                                    X_yoke_2,
                                                    Y_core_1,
                                                    Y_void_1,
                                                    Y_yoke_1,
                                                    Y_core_2,
                                                    Y_void_2,
                                                    Y_yoke_2,
                                                    Z_len, 
                                                    Z_pos=Z_pos,
                                                    lc=lc,
                                                    lc_inner=0.3*lc,
                                                    yoke_type=3)

   # the iron domain
   vol_air = gmsh.model.occ.addBox(0.0, 0.0, z_min, lim_x, lim_y, z_max - z_min)
   gmsh.model.occ.synchronize()

   # fragment perfoms something like a union
   fragments, _ = gmsh.model.occ.fragment([(3, vol_iron)], [(3, vol_air)])
   gmsh.model.occ.synchronize()

   # we get the domains of the fragmentation
   dom_iron = fragments[0][1]
   dom_air = fragments[1][1]

   # and we define physical domains
   gmsh.model.addPhysicalGroup(3, [dom_iron], 1, name = "Iron")
   gmsh.model.occ.synchronize()
   gmsh.model.addPhysicalGroup(3, [dom_air], 2, name = "Air")
   gmsh.model.occ.synchronize()

   # we then generate the mesh
   gmsh.model.mesh.generate(3)
   gmsh.model.occ.synchronize()

   gmsh.model.occ.synchronize()

   # ====================================================
   # Make the coil objects

   # this list stores the coil objects
   coil_list = []

   # determine the slot size
   slot_size = 2*min(Y_core_1, Y_core_2)

   # determine the number of conductors
   num_cond = np.int32(slot_size/2/(coil_radius+ins))

   # these are the vertical positions
   y = np.linspace(-0.5*slot_size + coil_radius + ins,
                    0.5*slot_size - coil_radius - ins, num_cond)

   # make two coils

   kp_1 = np.array([[ X_void_2 - yoke_spacer - ins, Z_pos + Z_len           ],
                  [ X_void_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_yoke_2,          Z_pos + Z_len + yoke_spacer + ins    ],
                  [ X_yoke_2 + yoke_spacer + ins,   Z_pos + Z_len           ],
                  [ X_yoke_1 + yoke_spacer + ins,   Z_pos                   ],
                  [ X_yoke_1,                       Z_pos-yoke_spacer - ins ],
                  [ X_void_1,                       Z_pos-yoke_spacer - ins ],
                  [ X_void_1 - yoke_spacer - ins,   Z_pos                   ]])
      
   kp_2 = kp_1.copy()
   kp_2[:, 0] *= -1.0
      
   coil_list.append(snoopy.RacetrackCoil(kp_1, y, coil_radius, current/num_cond))
   coil_list.append(snoopy.RacetrackCoil(kp_2, y, coil_radius, current/num_cond))


   snoopy.plot_domain(pl, gmsh.model.mesh, dom_iron, reflect_xz=True, reflect_yz=True, show_edges=show_edges, opacity=opacity, plot_feature_edges=plot_feature_edges)
   # snoopy.plot_domain(pl, gmsh.model.mesh, dom_air, reflect_xz=True, reflect_yz=True, show_edges=False, opacity=0.0)
   for coil in coil_list:
      coil.plot_pv(pl)


import numpy as np
import gmsh
from tqdm import tqdm
from scipy.sparse import csr_array
import matplotlib.pyplot as plt
import pandas as pd
from scipy.interpolate import BSpline
import pyvista as pv

from . import fem_c_mod as fem_c



def compute_2D_line_intersection(intercept_1, slope_1, intercept_2, slope_2):
    '''Compute the intersection point of two lines in 2D.

    :param intercept_1:
        The two coordinates of the interception point of line 1.

    :param slope_1:
        The two coordinates of the slope of line 1.

    :param intercept_2:
        The two coordinates of the interception point of line 2.

    :param slope_2:
        The two coordinates of the slope of line 2.
        
    :return:
        The two coordinates of the interception point.
    '''
    M = np.array([[slope_1[0] , -slope_2[0]],
                  [slope_1[1], -slope_2[1]]])

    b = np.array([intercept_2[0] - intercept_1[0],
                  intercept_2[1] - intercept_1[1]])

    x_2 = np.linalg.solve(M, b)

    return slope_1*x_2[0] + intercept_1


class RacetrackCoil():

    def __init__(self, kp, y, rad, current=1.0):
        '''Default constructor.
        
        :param kp:
            The 8 keypoints in the x,z plane.

        :param y:
            An array with vertical coordinates. There is one turn on
            each vertical coordinate.

        :param rad:
            The coil radius.

        :param current:
            The coil current.

        :return:
            None.
        '''

        self.kp = kp
        self.y = y
        self.rad = rad
        self.current = current

        return None

    def compute_B(self, positions, disc_arc=5):
        '''Compute the B field at given positions.
        
        :param positions:
            The points to evaluate.
            
        :param disc_arc:
            The discretization of the arcs into this number of segments.
            Default = 5.
            
        :return:
            The B field vectors.
            
        '''
        segs = self.get_segments(disc_arc=disc_arc)

        return fem_c.compute_B_line_segs_cpp(segs, positions, self.current, self.rad)

    def get_segments(self, disc_arc=5, enable_plot=False):
        '''Get the segments for this racetrack wire
        
        :param disc_arc:
            Optional parameter. The arcs are discretized according to this
            setting.

        :param enable_plot:
            Optional parameter. To enable a plot.

        :return:
            An (M x 6) numpy array where M is the number of segments.
        '''

        # we will use quadratic BSplines for the arcs
        k = 2

        # this is the knot vector
        t = [0, 0, 0, 1, 1, 1]

        # these will be the coefficients
        cx = [0., 0., 0.]
        cz = [0., 0., 0.]

        # this is a parameter vector we use to discretize the splines
        t_vec = np.linspace(0., 1., disc_arc+1)

        # on each level we have 4*disc_arc + 4 segments
        segs_per_lvl = 4*disc_arc + 4

        # make the argments on the xz plane
        segs_base = np.zeros((segs_per_lvl, 6))

        # this is the 'front' segment
        segs_base[0, :3] = np.array([self.kp[7, 0], 0.0, self.kp[7, 1]])
        segs_base[0, 3:] = np.array([self.kp[0, 0], 0.0, self.kp[0, 1]])

        # we determine the control point of the first arc from the intersection
        # of two lines       
        intersect_1 = compute_2D_line_intersection(self.kp[7, :],
                                                    self.kp[0, :] - self.kp[7, :],
                                                    self.kp[1, :],
                                                    self.kp[2, :] - self.kp[1, :])

        # make the first arc
        cx = [self.kp[0, 0], intersect_1[0], self.kp[1, 0]]
        cz = [self.kp[0, 1], intersect_1[1], self.kp[1, 1]]
        
        spl_x = BSpline(t, cx, k)(t_vec)
        spl_z = BSpline(t, cz, k)(t_vec)

        for i in range(disc_arc):
            segs_base[1+i, :3] = np.array([spl_x[i], 0.0, spl_z[i]])
            segs_base[1+i, 3:] = np.array([spl_x[i+1], 0.0, spl_z[i+1]])

        # this is the next straight segment
        segs_base[disc_arc+1, :3] = np.array([self.kp[1, 0], 0.0, self.kp[1, 1]])
        segs_base[disc_arc+1, 3:] = np.array([self.kp[2, 0], 0.0, self.kp[2, 1]])


        # we determine the control point of the first arc from the intersection
        # of two lines
        intersect_2 = compute_2D_line_intersection(self.kp[1, :],
                                                    self.kp[2, :] - self.kp[1, :],
                                                    self.kp[3, :],
                                                    self.kp[4, :] - self.kp[3, :])

        # make the first arc
        cx = [self.kp[2, 0], intersect_2[0], self.kp[3, 0]]
        cz = [self.kp[2, 1], intersect_2[1], self.kp[3, 1]]
        
        spl_x = BSpline(t, cx, k)(t_vec)
        spl_z = BSpline(t, cz, k)(t_vec)

        for i in range(disc_arc):
            segs_base[2+disc_arc+i, :3] = np.array([spl_x[i], 0.0, spl_z[i]])
            segs_base[2+disc_arc+i, 3:] = np.array([spl_x[i+1], 0.0, spl_z[i+1]])

        # this is the next straight segment
        segs_base[2+2*disc_arc, :3] = np.array([self.kp[3, 0], 0.0, self.kp[3, 1]])
        segs_base[2+2*disc_arc, 3:] = np.array([self.kp[4, 0], 0.0, self.kp[4, 1]])

        # we determine the control point of the first arc from the intersection
        # of two lines
        intersect_3 = compute_2D_line_intersection(self.kp[3, :],
                                                    self.kp[4, :] - self.kp[3, :],
                                                    self.kp[5, :],
                                                    self.kp[6, :] - self.kp[5, :])

        # make the first arc
        cx = [self.kp[4, 0], intersect_3[0], self.kp[5, 0]]
        cz = [self.kp[4, 1], intersect_3[1], self.kp[5, 1]]
        
        spl_x = BSpline(t, cx, k)(t_vec)
        spl_z = BSpline(t, cz, k)(t_vec)

        for i in range(disc_arc):
            segs_base[3+2*disc_arc+i, :3] = np.array([spl_x[i], 0.0, spl_z[i]])
            segs_base[3+2*disc_arc+i, 3:] = np.array([spl_x[i+1], 0.0, spl_z[i+1]])

        # this is the next straight segment
        segs_base[3+3*disc_arc, :3] = np.array([self.kp[5, 0], 0.0, self.kp[5, 1]])
        segs_base[3+3*disc_arc, 3:] = np.array([self.kp[6, 0], 0.0, self.kp[6, 1]])

        # we determine the control point of the first arc from the intersection
        # of two lines
        intersect_4 = compute_2D_line_intersection(self.kp[5, :],
                                                    self.kp[6, :] - self.kp[5, :],
                                                    self.kp[7, :],
                                                    self.kp[0, :] - self.kp[7, :])

        # make the first arc
        cx = [self.kp[6, 0], intersect_4[0], self.kp[7, 0]]
        cz = [self.kp[6, 1], intersect_4[1], self.kp[7, 1]]
        
        spl_x = BSpline(t, cx, k)(t_vec)
        spl_z = BSpline(t, cz, k)(t_vec)

        for i in range(disc_arc):
            segs_base[4+3*disc_arc+i, :3] = np.array([spl_x[i], 0.0, spl_z[i]])
            segs_base[4+3*disc_arc+i, 3:] = np.array([spl_x[i+1], 0.0, spl_z[i+1]])

        # the return data
        segs_ret = np.zeros((0, 6))

        # now stack vertically
        for i, yy in enumerate(self.y):
            segs_base[:, 1] = yy
            segs_base[:, 4] = yy
            segs_ret = np.append(segs_ret, segs_base, axis=0)

        if enable_plot:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            
            for i, seg in enumerate(segs_ret):
                ax.plot([seg[0], seg[3]],
                        [seg[1], seg[4]],
                        [seg[2], seg[5]])
            
            for i, yy in enumerate(self.y):
                ax.plot(intersect_1[0], yy, intersect_1[1], 'o', color='red')
                ax.plot(intersect_2[0], yy, intersect_2[1], 'o', color='red')
                ax.plot(intersect_3[0], yy, intersect_3[1], 'o', color='red')
                ax.plot(intersect_4[0], yy, intersect_4[1], 'o', color='red')
            ax.set_xlabel('$x$ in m')
            ax.set_xlabel('$y$ in m')
            ax.set_xlabel('$z$ in m')
            ax.set_aspect('equal')
            
            plt.show()

        return segs_ret
    

    def plot_mpl(self, ax):
        '''Plot the geometry in a matplotlib axes object.

        :param ax:
            A matplotlib axes object.

        :return:
            None.
        '''

        segs = self.get_segments()
            
        for i, seg in enumerate(segs):
            ax.plot([seg[0], seg[3]],
                    [seg[1], seg[4]],
                    [seg[2], seg[5]])
            
        return
    
    def plot_pv(self, pl, color=[255/255, 40/255, 0/255], width=8):
        '''Plot the geometry in a pyvista plotter object.

        :param pl:
            A pyvista plotter object.

        :return:
            None.
        '''

        segs = self.get_segments()

        points, cells = self.make_surface_mesh(disc_arc=5)


        # make the mesh object
        cell_info = np.append(np.ones((len(cells), 1), dtype=np.int64)*4, np.int64(cells), axis=1)
        mesh = pv.UnstructuredGrid(cell_info, [pv.CellType.QUAD]*cells.shape[0], points)

        # pl.add_mesh(mesh, color=color)
        surf = mesh.extract_surface()
        surf.compute_normals(inplace=True, split_vertices=True)
        pl.add_mesh(surf, color=color, pbr=True, metallic=0.5, roughness=0.3)

        
        return

    def make_surface_mesh(self, resol=30, disc_arc=5):
        '''Mesh the surface of the cable.

        :param resol:
            The discretization in the azimutal direction. Default 12.

        :return:
            The nodes and the connectivity of the mesh.
        '''

        # get the segments
        segs = self.get_segments(disc_arc=disc_arc)

        # the number of segments
        num_seg = segs.shape[0]

        # number of vertical positions
        num_y = len(self.y)

        # the number of segments per level
        num_segs_per_level = np.int32(num_seg/num_y)

        # total number of points
        num_points = (num_segs_per_level+1)*resol

        # points container
        points = np.zeros((num_points, 3))

        # number of cells
        num_cells = num_segs_per_level*(resol-1)

        # cells container
        cells = np.zeros((num_cells, 4), dtype=np.int32)

        # the azimuth angles
        phi = np.linspace(-np.pi, np.pi, resol)

        # fill the points
        for i, seg in enumerate(segs[:num_segs_per_level, :]):

            # the direction vector
            d = (seg[3:] - seg[:3])/np.linalg.norm(seg[3:] - seg[:3])

            # the transversal unit vectors
            e_u = np.cross(d, np.array([0., 1., 0.]))

            for j, pphi in enumerate(phi):

                points[i*resol + j, 0] = seg[0] + self.rad*np.cos(pphi)*e_u[0]
                points[i*resol + j, 1] = seg[1] + self.rad*np.sin(pphi)
                points[i*resol + j, 2] = seg[2] + self.rad*np.cos(pphi)*e_u[2]

            if i == num_segs_per_level - 1:
                for j, pphi in enumerate(phi):

                    points[(i+1)*resol + j, 0] = seg[3] + self.rad*np.cos(pphi)*e_u[0]
                    points[(i+1)*resol + j, 1] = seg[4] + self.rad*np.sin(pphi)
                    points[(i+1)*resol + j, 2] = seg[5] + self.rad*np.cos(pphi)*e_u[2]



        # fill the cells
        for i in range(num_segs_per_level):
            for j in range(resol - 1):
                cells[i*(resol-1) + j, :] = np.array([i*resol + j,
                                                     (i+1)*resol + j,
                                                      (i+1)*resol + j + 1,
                                                       i*resol + j + 1], dtype=np.int32)

            
        # all points
        all_points = np.zeros((len(self.y)*points.shape[0], 3))

        # all cells
        all_cells = np.zeros((len(self.y)*cells.shape[0], 4), dtype=np.int32)

        # stack the coils again
        for i, yy in enumerate(self.y):
            
            all_points[i*points.shape[0]:(i+1)*points.shape[0], :] = points
            all_points[i*points.shape[0]:(i+1)*points.shape[0], 1] += yy - self.y[0]

            all_cells[i*cells.shape[0]:(i+1)*cells.shape[0], :] = cells + i*points.shape[0]

        return all_points, all_cells



import numpy as np
import gmsh
import pyvista as pv

def add_SHIP_coil(model, X_core_1, X_void_1, X_yoke_1,
                    X_core_2, X_void_2, X_yoke_2,
                    Z_len, W_coil, H_coil,
                    Z_pos=0.0,
                    show=False,
                    lc=1.0,
                    coil_type=1,
                    sym_xz=True):
    '''Add a coil domain to a gmsh model for SHiP.

    :param model:
        The gmsh model.

    :param X_core_1:
        The horizontal coordinate of the end of the iron core (entrance).

    :param X_void_1:
        The horizontal coordinate of the end of the void region (entrance).

    :param X_yoke_1:
        The horizontal coordinate of the end of the magnet end (entrance).

    :param X_core_2:
        The horizontal coordinate of the end of the iron core (exit).

    :param X_void_2:
        The horizontal coordinate of the end of the void region (exit).

    :param X_yoke_2:
        The horizontal coordinate of the end of the magnet end (exit).

    :param Z_len:
        The length of the magnet in the z direction.

    :param W_coil:
        The coil width.

    :param H_coil:
        The coil height,

    :param Z_pos:
        The position of the magnet in the z direction.

    :param show:
        Set this flag to show the mesh in the gmsh gui.

    :param coil_type:
        The coil type. Type 1 is a coil that is symmetric in the yz plane.

    :param sym_xz:
        Set this flag to true if You want to apply symmetry in the xz plane.

    :return:
        The volume tag of the coil.
    '''

    # the tapering angle alpha
    alpha = np.arctan2(X_core_2 - X_core_1, Z_len)

    # the z position of the entrance
    Z_1 = Z_pos

    # the z position of the exit
    Z_2 = Z_1 + Z_len

    # make the key points
    if sym_xz:
        p1 = model.occ.addPoint(0.0, 0.0, Z_1, lc)
        p2 = model.occ.addPoint(0.0, 0.0, Z_1 - W_coil, lc)
        p3 = model.occ.addPoint(X_core_1, 0.0, Z_1 - W_coil, lc)
        p4 = model.occ.addPoint(X_core_1 + W_coil*np.cos(alpha), 0.0, Z_1 - W_coil*np.sin(alpha), lc)
        p5 = model.occ.addPoint(X_core_2 + W_coil*np.cos(alpha), 0.0, Z_2 +  W_coil*np.sin(alpha), lc)
        p6 = model.occ.addPoint(X_core_2, 0.0, Z_2 + W_coil, lc)
        p7 = model.occ.addPoint(0.0, 0.0, Z_2 + W_coil, lc)
        p8 = model.occ.addPoint(0.0, 0.0, Z_2, lc)
        p9 = model.occ.addPoint(X_core_2, 0.0, Z_2, lc)
        p10 = model.occ.addPoint(X_core_1, 0.0, Z_1, lc)

    else:
        p1 = model.occ.addPoint(0.0, -H_coil , Z_1, lc)
        p2 = model.occ.addPoint(0.0, -H_coil, Z_1 - W_coil, lc)
        p3 = model.occ.addPoint(X_core_1, -H_coil, Z_1 - W_coil, lc)
        p4 = model.occ.addPoint(X_core_1 + W_coil*np.cos(alpha), -H_coil, Z_1 - W_coil*np.sin(alpha), lc)
        p5 = model.occ.addPoint(X_core_2 + W_coil*np.cos(alpha), -H_coil, Z_2 +  W_coil*np.sin(alpha), lc)
        p6 = model.occ.addPoint(X_core_2, -H_coil, Z_2 + W_coil, lc)
        p7 = model.occ.addPoint(0.0, -H_coil, Z_2 + W_coil, lc)
        p8 = model.occ.addPoint(0.0, -H_coil, Z_2, lc)
        p9 = model.occ.addPoint(X_core_2, -H_coil, Z_2, lc)
        p10 = model.occ.addPoint(X_core_1, -H_coil, Z_1, lc)

    # make the lines
    l1 = model.occ.addLine(p1, p2)
    l2 = model.occ.addLine(p2, p3)
    l3 = model.occ.addCircleArc(p3, p10, p4)
    l4 = model.occ.addLine(p4, p5)
    l5 = model.occ.addCircleArc(p5, p9, p6)
    l6 = model.occ.addLine(p6, p7)
    l7 = model.occ.addLine(p7, p8)
    l8 = model.occ.addLine(p8, p9)
    l9 = model.occ.addLine(p9, p10)
    l10 = model.occ.addLine(p10, p1)

    # make a curve loop
    c = model.occ.addCurveLoop([l1, l2, l3, l4, l5, l6, l7, l8, l9, l10])

    # make the surfaces
    s = model.occ.addPlaneSurface([c])

    # extrude
    if sym_xz:
        extr = model.occ.extrude ([(2, s)], 0., H_coil, 0.)
    else:
        extr = model.occ.extrude ([(2, s)], 0., 2*H_coil, 0.)

    model.occ.synchronize()

    if show:
        gmsh.fltk.run()

    # get the extruded volume
    vol = -1
    for e in extr:
        if e[0] == 3:
            vol = e[1]
            break

    return vol

def add_SHIP_iron_yoke(model, X_mgap_1, X_core_1, X_void_1, X_yoke_1,
                        X_mgap_2, X_core_2, X_void_2, X_yoke_2,
                        Y_core_1, Y_void_1, Y_yoke_1,
                        Y_core_2, Y_void_2, Y_yoke_2,
                        Z_len, 
                        Z_pos=0.0,
                        lc=2e-1,
                        lc_inner=-1,
                        yoke_type=1,
                        reflect_xz=False):
    '''Add an iron yoke for the SHiP project to a gmsh model.
    
    :param X_mgap_1:
        The size of the horizontal gap. (entrance)

    :param X_core_1:
        The horizontal coordinate of the end of the iron core (entrance).

    :param X_void_1:
        The horizontal coordinate of the end of the void region (entrance).

    :param X_yoke_1:
        The horizontal coordinate of the end of the magnet end (entrance).

    :param X_mgap_2:
        The size of the horizontal gap. (exit)

    :param X_core_2:
        The horizontal coordinate of the end of the iron core (exit).

    :param X_void_2:
        The horizontal coordinate of the end of the void region (exit).

    :param X_yoke_2:
        The horizontal coordinate of the end of the magnet end (exit).

    :param Y_void_1:
        The vertical coordinate of the end of the void region. (entrance)

    :param Y_yoke_1:
        The vertical coordinate of the end of the iron yoke (entrance).

    :param Y_void_2:
        The vertical coordinate of the end of the void region (exit).

    :param Y_yoke_2:
        The horizontal coordinate of the end of the iron yoke (exit).

    :param Z_len:
        The length of the magnet in the z direction.

    :param Z_pos:
        The position of the magnet in the z direction.

    :param show:
        Set this flag to show the mesh in the gmsh gui.

    :param lc:
        A length scale parameter to control the mesh size.

    :param lc_inner:
        The legth scale parameter applied in the inner of the iron domain.
        If negative it is ignored.

    :param yoke_type:
        The yoke type. Options are 1, 2, and 3.
        The types 1 and 3 are identical (magnet templates for normal
        conducting magnets)
        The type 2 is the one for the SC magnet.

    :param reflect_xz:
        Reflect the fomain in the xz plane.

    :return:
        The gmsh model object as well as the labels of the physical groups for the two terminals.
    '''

    if lc_inner < 0:
        lc_inner = lc

    if (yoke_type == 1 or yoke_type == 3):

        # the z position of the entrance
        Z_1 = Z_pos

        # the z position of the exit
        Z_2 = Z_1 + Z_len
        
        # THIS IS THE CORE

        # make the points in the xy plane (entrance)
        p1_1 = model.occ.addPoint(X_mgap_1, 0.0, Z_1, lc)
        p2_1 = model.occ.addPoint(X_core_1, 0.0, Z_1, lc)
        p3_1 = model.occ.addPoint(X_core_1, Y_core_1, Z_1, lc)
        p4_1 = model.occ.addPoint(X_void_1, Y_core_1, Z_1, lc)
        p5_1 = model.occ.addPoint(X_void_1, 0.0, Z_1, lc)
        p6_1 = model.occ.addPoint(X_yoke_1, 0.0, Z_1, lc)
        p7_1 = model.occ.addPoint(X_yoke_1, Y_yoke_1, Z_1, lc)
        p8_1 = model.occ.addPoint(X_mgap_1, Y_yoke_1, Z_1, lc)

        # connect the keypoints by lines (entrance)
        l1_1 = model.occ.addLine(p1_1, p2_1)
        l2_1 = model.occ.addLine(p2_1, p3_1)
        l3_1 = model.occ.addLine(p3_1, p4_1)
        l4_1 = model.occ.addLine(p4_1, p5_1)
        l5_1 = model.occ.addLine(p5_1, p6_1)
        l6_1 = model.occ.addLine(p6_1, p7_1)
        l7_1 = model.occ.addLine(p7_1, p8_1)
        l8_1 = model.occ.addLine(p8_1, p1_1)

        # make the points in the xy plane (exit)
        p1_2 = model.occ.addPoint(X_mgap_2, 0.0, Z_2, lc)
        p2_2 = model.occ.addPoint(X_core_2, 0.0, Z_2, lc)
        p3_2 = model.occ.addPoint(X_core_2, Y_core_2, Z_2, lc)
        p4_2 = model.occ.addPoint(X_void_2, Y_core_2, Z_2, lc)
        p5_2 = model.occ.addPoint(X_void_2, 0.0, Z_2, lc)
        p6_2 = model.occ.addPoint(X_yoke_2, 0.0, Z_2, lc)
        p7_2 = model.occ.addPoint(X_yoke_2, Y_yoke_2, Z_2, lc)
        p8_2 = model.occ.addPoint(X_mgap_2, Y_yoke_2, Z_2, lc)

        # connect the keypoints by lines (exit)
        l1_2 = model.occ.addLine(p1_2, p2_2)
        l2_2 = model.occ.addLine(p2_2, p3_2)
        l3_2 = model.occ.addLine(p3_2, p4_2)
        l4_2 = model.occ.addLine(p4_2, p5_2)
        l5_2 = model.occ.addLine(p5_2, p6_2)
        l6_2 = model.occ.addLine(p6_2, p7_2)
        l7_2 = model.occ.addLine(p7_2, p8_2)
        l8_2 = model.occ.addLine(p8_2, p1_2)

        # connect the entrance and exit by lines
        l1_12 = model.occ.addLine(p1_1, p1_2)
        l2_12 = model.occ.addLine(p2_1, p2_2)
        l3_12 = model.occ.addLine(p3_1, p3_2)
        l4_12 = model.occ.addLine(p4_1, p4_2)
        l5_12 = model.occ.addLine(p5_1, p5_2)
        l6_12 = model.occ.addLine(p6_1, p6_2)
        l7_12 = model.occ.addLine(p7_1, p7_2)
        l8_12 = model.occ.addLine(p8_1, p8_2)
        
        # make all curve loops
        c1 = model.occ.addCurveLoop([l1_1, l2_1, l3_1, l4_1, l5_1, l6_1, l7_1, l8_1])
        c2 = model.occ.addCurveLoop([l1_2, l2_2, l3_2, l4_2, l5_2, l6_2, l7_2, l8_2])
        c3 = model.occ.addCurveLoop([l1_1, l2_12, l1_2, l1_12])
        c4 = model.occ.addCurveLoop([l2_1, l3_12, l2_2, l2_12])
        c5 = model.occ.addCurveLoop([l3_1, l4_12, l3_2, l3_12])

        c6 = model.occ.addCurveLoop([l4_1, l5_12, l4_2, l4_12])
        c7 = model.occ.addCurveLoop([l5_1, l6_12, l5_2, l5_12])
        c8 = model.occ.addCurveLoop([l6_1, l7_12, l6_2, l6_12])
        c9 = model.occ.addCurveLoop([l7_1, l8_12, l7_2, l7_12])
        c10 = model.occ.addCurveLoop([l8_1, l1_12, l8_2, l8_12])

        # make the surfaces
        s1 = model.occ.addPlaneSurface([c1])
        s2 = model.occ.addPlaneSurface([c2])
        s3 = model.occ.addPlaneSurface([c3])
        s4 = model.occ.addPlaneSurface([c4])
        s5 = model.occ.addPlaneSurface([c5])
        s6 = model.occ.addPlaneSurface([c6])
        s7 = model.occ.addPlaneSurface([c7])
        s8 = model.occ.addPlaneSurface([c8])
        s9 = model.occ.addPlaneSurface([c9])
        s10 = model.occ.addPlaneSurface([c10])

        # make the iron and air domains
        sl_1 = model.occ.addSurfaceLoop([s1, s2, s3, s4, s5, s6, s7, s8, s9, s10])
        vol_1 = model.occ.addVolume([sl_1])

        model.occ.synchronize()

    elif yoke_type == 2:

        # the z position of the entrance
        Z_1 = Z_pos

        # the z position of the exit
        Z_2 = Z_1 + Z_len
        
        # THIS IS THE CORE

        # make the points in the xy plane (entrance)
        p1_1 = model.occ.addPoint(0.0, Y_void_1, Z_1, lc_inner)
        p2_1 = model.occ.addPoint(X_void_1, Y_void_1, Z_1, lc_inner)
        p3_1 = model.occ.addPoint(X_void_1, 0.0, Z_1, lc_inner)
        p4_1 = model.occ.addPoint(X_yoke_1, 0.0, Z_1, lc)
        p5_1 = model.occ.addPoint(X_yoke_1, Y_yoke_1, Z_1, lc)
        p6_1 = model.occ.addPoint(0.0, Y_yoke_1, Z_1, lc)

        # connect the keypoints by lines (entrance)
        l1_1 = model.occ.addLine(p1_1, p2_1)
        l2_1 = model.occ.addLine(p2_1, p3_1)
        l3_1 = model.occ.addLine(p3_1, p4_1)
        l4_1 = model.occ.addLine(p4_1, p5_1)
        l5_1 = model.occ.addLine(p5_1, p6_1)
        l6_1 = model.occ.addLine(p6_1, p1_1)

        # make the points in the xy plane (exit)
        p1_2 = model.occ.addPoint(0.0, Y_void_2, Z_2, lc_inner)
        p2_2 = model.occ.addPoint(X_void_2, Y_void_2, Z_2, lc_inner)
        p3_2 = model.occ.addPoint(X_void_2, 0.0, Z_2, lc_inner)
        p4_2 = model.occ.addPoint(X_yoke_2, 0.0, Z_2, lc)
        p5_2 = model.occ.addPoint(X_yoke_2, Y_yoke_2, Z_2, lc)
        p6_2 = model.occ.addPoint(0.0, Y_yoke_2, Z_2, lc)

        # connect the keypoints by lines (exit)
        l1_2 = model.occ.addLine(p1_2, p2_2)
        l2_2 = model.occ.addLine(p2_2, p3_2)
        l3_2 = model.occ.addLine(p3_2, p4_2)
        l4_2 = model.occ.addLine(p4_2, p5_2)
        l5_2 = model.occ.addLine(p5_2, p6_2)
        l6_2 = model.occ.addLine(p6_2, p1_2)

        # connect the entrance and exit by lines
        l1_12 = model.occ.addLine(p1_1, p1_2)
        l2_12 = model.occ.addLine(p2_1, p2_2)
        l3_12 = model.occ.addLine(p3_1, p3_2)
        l4_12 = model.occ.addLine(p4_1, p4_2)
        l5_12 = model.occ.addLine(p5_1, p5_2)
        l6_12 = model.occ.addLine(p6_1, p6_2)
        
        # make all curve loops
        c1 = model.occ.addCurveLoop([l1_1, l2_1, l3_1, l4_1, l5_1, l6_1])
        c2 = model.occ.addCurveLoop([l1_2, l2_2, l3_2, l4_2, l5_2, l6_2])
        c3 = model.occ.addCurveLoop([l1_1, l2_12, l1_2, l1_12])
        c4 = model.occ.addCurveLoop([l2_1, l3_12, l2_2, l2_12])
        c5 = model.occ.addCurveLoop([l3_1, l4_12, l3_2, l3_12])

        c6 = model.occ.addCurveLoop([l4_1, l5_12, l4_2, l4_12])
        c7 = model.occ.addCurveLoop([l5_1, l6_12, l5_2, l5_12])

        # make the surfaces
        s1 = model.occ.addPlaneSurface([c1])
        s2 = model.occ.addPlaneSurface([c2])
        s3 = model.occ.addPlaneSurface([c3])
        s4 = model.occ.addPlaneSurface([c4])
        s5 = model.occ.addPlaneSurface([c5])
        s6 = model.occ.addPlaneSurface([c6])
        s7 = model.occ.addPlaneSurface([c7])

        # make the iron and air domains
        sl_1 = model.occ.addSurfaceLoop([s1, s2, s3, s4, s5, s6, s7])
        vol_1 = model.occ.addVolume([sl_1])

        model.occ.synchronize()

    else:
        print('Yoke type {} is unknown!'.format(yoke_type))

    if reflect_xz:

        model.occ.mirror([(3, vol_1)], 0., 1., 0., 0.)
        model.occ.synchronize()
        # dim_tags, map = model.occ.fragment(vol_m, [(3, vol_1)])
        # model.occ.synchronize()

    return vol_1


def add_SHIP_iron_core(model, X_core_1, X_core_2, Y_core_1, Y_core_2,
                        Z_len, Z_pos=0.0, lc=2e-1, sym_xz=True):
    '''Add an iron core for the SHiP project to a gmsh model.
    
    :param X_core_1:
        The horizontal coordinate of the end of the iron core (entrance).

    :param X_core_2:
        The horizontal coordinate of the end of the iron core (exit).

    :param Y_core_1:
        The vertical coordinate of the end of the void region. (entrance)

    :param Y_core_2:
        The horizontal coordinate of the end of the iron yoke (exit).

    :param Z_len:
        The length of the magnet in the z direction.

    :param Z_pos:
        The position of the magnet in the z direction.

    :param show:
        Set this flag to show the mesh in the gmsh gui.

    :param lc:
        A length scale parameter to control the mesh size.

    :param sym_xz:
        Set this flag to true if You want to apply symmetry in the xz plane.
        
    :return:
        The gmsh model object as well as the labels of the physical groups for the two terminals.
    '''

    # the z position of the entrance
    Z_1 = Z_pos

    # the z position of the exit
    Z_2 = Z_1 + Z_len
        
    # THIS IS THE CORE

    if sym_xz:
        # make the points in the xy plane (entrance)
        p1_1 = model.occ.addPoint(0.0, 0.0, Z_1, lc)
        p2_1 = model.occ.addPoint(X_core_1, 0.0, Z_1, lc)
        p3_1 = model.occ.addPoint(X_core_1, Y_core_1, Z_1, lc)
        p4_1 = model.occ.addPoint(0.0, Y_core_1, Z_1, lc)
    else:
        # make the points in the xy plane (entrance)
        p1_1 = model.occ.addPoint(0.0, -Y_core_1, Z_1, lc)
        p2_1 = model.occ.addPoint(X_core_1, -Y_core_1, Z_1, lc)
        p3_1 = model.occ.addPoint(X_core_1, Y_core_1, Z_1, lc)
        p4_1 = model.occ.addPoint(0.0, Y_core_1, Z_1, lc)

    # connect the keypoints by lines (entrance)
    l1_1 = model.occ.addLine(p1_1, p2_1)
    l2_1 = model.occ.addLine(p2_1, p3_1)
    l3_1 = model.occ.addLine(p3_1, p4_1)
    l4_1 = model.occ.addLine(p4_1, p1_1)

    if sym_xz:
        # make the points in the xy plane (exit)
        p1_2 = model.occ.addPoint(0.0, 0.0, Z_2, lc)
        p2_2 = model.occ.addPoint(X_core_2, 0.0, Z_2, lc)
        p3_2 = model.occ.addPoint(X_core_2, Y_core_2, Z_2, lc)
        p4_2 = model.occ.addPoint(0.0, Y_core_2, Z_2, lc)
    else:
        # make the points in the xy plane (exit)
        p1_2 = model.occ.addPoint(0.0, -Y_core_2, Z_2, lc)
        p2_2 = model.occ.addPoint(X_core_2, -Y_core_2, Z_2, lc)
        p3_2 = model.occ.addPoint(X_core_2, Y_core_2, Z_2, lc)
        p4_2 = model.occ.addPoint(0.0, Y_core_2, Z_2, lc)

    # connect the keypoints by lines (exit)
    l1_2 = model.occ.addLine(p1_2, p2_2)
    l2_2 = model.occ.addLine(p2_2, p3_2)
    l3_2 = model.occ.addLine(p3_2, p4_2)
    l4_2 = model.occ.addLine(p4_2, p1_2)
    
    # connect the entrance and exit by lines
    l1_12 = model.occ.addLine(p1_1, p1_2)
    l2_12 = model.occ.addLine(p2_1, p2_2)
    l3_12 = model.occ.addLine(p3_1, p3_2)
    l4_12 = model.occ.addLine(p4_1, p4_2)
        
    # make all curve loops
    c1 = model.occ.addCurveLoop([l1_1, l2_1, l3_1, l4_1])
    c2 = model.occ.addCurveLoop([l1_2, l2_2, l3_2, l4_2])
    c3 = model.occ.addCurveLoop([l1_1, l2_12, l1_2, l1_12])
    c4 = model.occ.addCurveLoop([l2_1, l3_12, l2_2, l2_12])
    c5 = model.occ.addCurveLoop([l3_1, l4_12, l3_2, l3_12])
    c6 = model.occ.addCurveLoop([l4_1, l1_12, l4_2, l4_12])
    
    # make the surfaces
    s1 = model.occ.addPlaneSurface([c1])
    s2 = model.occ.addPlaneSurface([c2])
    s3 = model.occ.addPlaneSurface([c3])
    s4 = model.occ.addPlaneSurface([c4])
    s5 = model.occ.addPlaneSurface([c5])
    s6 = model.occ.addPlaneSurface([c6])
    
    # make the iron and air domains
    sl_1 = model.occ.addSurfaceLoop([s1, s2, s3, s4, s5, s6])
    vol_1 = model.occ.addVolume([sl_1])

    return vol_1


import numpy as np
import gmsh
import pyvista as pv

from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import cg
from scipy.sparse import diags

import matplotlib.pyplot as plt

from .laplace_subdomain_solver import LaplaceSubdomainSolver
from .curl_curl_assembler import CurlCurlAssembler
from .matrix_assembly import delete_rows_sparse
from .matrix_assembly import delete_cols_sparse
from .matrix_assembly import recover_full_solution

from .materials import ConstantReluctance

from .mesh_tools import get_edge_boundary_dofs
from .mesh_tools import get_cotree_dofs

from .plot_tools import plot_vector_field

class MVPSolver():

    def __init__(self, gmsh_model, coil_tags, iron_tags,
                 air_tags, coil_terminals, reluctances,
                 max_newton_iterations=25,
                 element_order=1,
                 quad_order=-1):
        '''Setup the solver with the geometry info.
        
        :param gmsh_model:
            The gmsh model object.
            
        :param coil_tags:
            The tags for the coil domains.

        :param iron_tags:
            The tags for the iron domains.

        :param air_tags:
            The tags for the air domains.

        :param coil_terminals:
            A list of tuples for the tags of the coil terminals. In the same order
            than the coil_tags.

        :param reluctances:
            A list with reluctances for the iron domains. In the same order
            than the iron_tags.
        :param element_order:
            The order of the FEM approximation space for the vector potential.

        :param max_newton_iteration:
            The maximum number of Newton iterations.

        :return:
            None
        '''

        # initialize gmsh if not done already
        if not gmsh.isInitialized():
            gmsh.initialize()

        # initialize the subdomain solvers for the coil domains
        self.coil_sd_solvers = []
        for i, ct in enumerate(coil_tags):
            self.coil_sd_solvers.append(LaplaceSubdomainSolver(gmsh_model.mesh, ct, coil_terminals[i]))

        # append all domain tags for the vector potential solver
        domain_tags = coil_tags + iron_tags + air_tags

        # append all material properties
        material_properties = [ConstantReluctance(0.25/np.pi*1e7)]*len(coil_tags) + reluctances + [ConstantReluctance(0.25/np.pi*1e7)]*len(air_tags)

        # the maximum number of Newton iterations
        self.max_newton_iterations = max_newton_iterations

        # store the finite element order
        self.element_order = element_order

        if (element_order > 3):
            print('Error! Element order {} not implemented! Choosing 1.')
            element_order = 1

        # store the quadrature order
        if quad_order < 0:
            self.quad_order = element_order + 2
        else:
            self.quad_order = quad_order

        # get the dirichlet boundary dofs
        self.dirichlet_dofs = get_edge_boundary_dofs(gmsh_model, ["Dirichlet Boundary"], element_order)

        # setup the matrix factory for the vector potential solver
        self.curl_curl_factory = CurlCurlAssembler(gmsh_model.mesh, domain_tags, material_properties, element_order)

        # the number of dofs for the vector potential
        self.num_dofs_mvp = self.curl_curl_factory.num_dofs


    def get_global_ids(self):
        '''Get the global ids for the field solution.

        :return:
            A (E x N) matrix (int) where E is the number of elements and N is the number
            of edge basis functions per element. 
        '''
        return self.curl_curl_factory.global_ids

    def make_boundary_mask(self):
        '''Make a mask to mask out the boundary degrees of freedom.

        :return:
            The boundary mask.
        '''

        # make the mask
        mask = np.ones((self.num_dofs_mvp, ), dtype=np.bool)

        # set the boundary dofs to fals
        for bc in self.dirichlet_dofs:
            mask[bc-1] = False

        return mask

    def solve_lin(self, phi_list, tolerance = 1e-4, xcg_0 = np.zeros((0, )), maxiter=-1, apply_gauge=False):
        '''Solve the problem.
        
        :param phi_list:
            The potential differences at the terminals to drive the current.

        :param tolerance:
            The tolerance for the cg iterations.

        :param xcg_0:
            The initial guess for the cg iterations.
            
        :param maxiter:
            The maximum number of iterations. If -1, the default scipy setting is used.

        :return:
            The solution vector.
        '''

        # a list with all scalar potential solution vectors
        sol_sp = []

        # the right hand side for the curl curl equation
        rhs_mvp = np.zeros((self.num_dofs_mvp, ))

        # the initial solution vector for the mvp here 0, a linear problem
        x_0 = np.zeros((self.num_dofs_mvp, ))

        # loop over all coil domain solvers
        for i, sol in enumerate(self.coil_sd_solvers):

            # solve for the electric scalar potential in the subdomain
            sol_sp.append(sol.solve(phi_list[i]))

            # increment the right hand side
            rhs_mvp += self.curl_curl_factory.compute_rhs_electric_potential(sol_sp[-1],
                                                                             quad_order=self.quad_order,
                                                                             select_mat=[i])

        # compute the contribution to the rhs from this coil
        K, _ = self.curl_curl_factory.compute_stiffness_and_jacobi_matrix_c(x_0, quad_order=self.quad_order)

        # make a boundary mask for the dirichlet boundary condition
        mask = self.make_boundary_mask()

        if apply_gauge:
            # get the spanning tree
            tree_dofs, cotree_dofs = get_cotree_dofs(self.curl_curl_factory.mesh)
            mask[tree_dofs] = False

        # apply the boundary condition
        K = K[mask, :]
        K = K[:, mask]
        rhs_mvp = rhs_mvp[mask]

        if len(xcg_0) == len(mask):
            xcg_0 = xcg_0[mask]

        # jacobi preconditioner
        M = diags(1.0/K.diagonal(), shape=K.shape)

        fig = plt.figure()
        ax = fig.add_subplot(121)
        ax.plot(rhs_mvp)
        ax = fig.add_subplot(122)
        cntrf = ax.contourf(K.toarray())
        fig.colorbar(cntrf)
        plt.show()

        # solve the problem
        # print("solve by conjugare gradients")
        if maxiter < 0:
            if len(xcg_0) == len(rhs_mvp):
                x, exit_code = cg(K, rhs_mvp, atol=tolerance, x0=xcg_0, M=M)
            else:
                x, exit_code = cg(K, rhs_mvp, M=M)
        else:
            if len(xcg_0) == len(rhs_mvp):
                x, exit_code = cg(K, rhs_mvp, atol=tolerance, x0=xcg_0, M=M, maxiter=maxiter)
            else:
                x, exit_code = cg(K, rhs_mvp, atol=tolerance, M=M, maxiter=maxiter)

        # we need to get back to the full solution vector
        x_long = 0.0*x_0
        x_long[mask == True] = x

        print('exit code = {}'.format(exit_code))
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(x)
        plt.show()

        return x_long

    def solve_nonlin(self, phi_list, tolerance = 1e-4, x_0=np.zeros((0, )), maxiter=-1, apply_gauge=False, plot_current_density=False):
            '''Solve the problem.
            
            :param phi_list:
                The potential differences at the terminals to drive the current.

            :param tolerance:
                The tolerance for the cg iterations.

            :param xcg_0:
                The initial guess for the cg iterations.
                
            :param maxiter:
                The maximum number of iterations. If -1, the default scipy setting is used.

            :param apply_gauge:
                Set this flag true to enable the tree-cotree gauged formulation.

            :return:
                The solution vector.
            '''

            # a list with all scalar potential solution vectors
            sol_sp = []

            # the right hand side for the curl curl equation
            rhs_mvp = np.zeros((self.num_dofs_mvp, ))

            # the initial solution vector for the mvp here 0, a linear problem
            if len(x_0) == 0:
                x_0 = np.zeros((self.num_dofs_mvp, ))

            # make a boundary mask for the dirichlet boundary condition
            mask = self.make_boundary_mask()

            if apply_gauge:
                # get the spanning tree
                tree_dofs, cotree_dofs = get_cotree_dofs(self.curl_curl_factory.mesh)
                mask[tree_dofs] = False

            # loop over all coil domain solvers
            for i, sol in enumerate(self.coil_sd_solvers):

                # solve for the electric scalar potential in the subdomain
                sol_sp.append(sol.solve(phi_list[i]))

                # increment the right hand side
                rhs_mvp += self.curl_curl_factory.compute_rhs_electric_potential(sol_sp[-1],
                                                                                 quad_order=self.quad_order,
                                                                                 select_mat=[i])

            if plot_current_density:
                
                pl = pv.Plotter()

                for i, x_sp in enumerate(sol_sp):
                    pp, JJ = self.coil_sd_solvers[i].matrix_factory.compute_field(x_sp)
                
                    plot_vector_field(pl, pp, JJ, title='J_{} in A/m**2'.format(i), mag=1/phi_list[i])
                
                pl.add_axes()
                pl.show_grid()
                pl.show()

            # reduce equation system (impose boundary conditions)
            rhs_mvp_c = rhs_mvp[mask]
            x_0_c = x_0[mask]

            # initalize current solution vector
            x_n = x_0.copy()

            
            print('***********************')
            print('start Newton iterations')
            for i in range(self.max_newton_iterations):
                
                # compute the contribution to the rhs from this coil
                K, J = self.curl_curl_factory.compute_stiffness_and_jacobi_matrix_c(x_n, quad_order=self.quad_order)

                # compute right hand side
                rhs =  J @ x_n - K @ x_n + rhs_mvp

                # reduce equation system (impose boundary conditions)
                J_c = J[mask, :]
                J_c = J_c[:, mask]

                # jacobi preconditioner
                M = diags(1.0/J_c.diagonal(), shape=J_c.shape)

                # launch cg
                if maxiter < 0:
                    xx, exit_code = cg(J_c, rhs[mask], atol=tolerance, x0=x_n[mask], M=M)
                else:
                    xx, exit_code = cg(J_c, rhs[mask], atol=tolerance, x0=x_n[mask], M=M, maxiter=maxiter)

                # we need to get back to the full solution vector
                x_np1 = 0.0*x_n
                x_np1[mask == True] = xx

                # the step
                h = x_np1 - x_n

                print('  step {}'.format(i))
                print('    max relative increment {:.3e}'.format(max(abs(h))/max(abs(x_np1))))

                # stop if tolerance is reached
                if max(abs(h)) < tolerance*max(abs(x_np1)):
                    print('  tolerance reached!')
                    break
                else:
                    x_n = x_np1.copy()

            return x_np1
